@@ -60,6 +60,17 @@
 #'   the overdispersion varies based on the mean expression level (lower expression level => higher
 #'   dispersion). If `overdispersion_shrinkage = TRUE`, a median trend of dispersion and expression level is
 #'   fit and used to estimate the variances of a quasi Gamma Poisson model (Lund et al. 2012). Default: `TRUE`.
+#' @param ridge_penalty to avoid overfitting, we can penalize fits with large coefficient estimates. Instead
+#'   of directly minimizing the deviance per gene (\eqn{Sum dev(y_i, X_i b)}), we will minimize
+#'   \eqn{Sum dev(y_i, X_i b) + N * Sum (penalty_p * b_p)^2}.\cr
+#'   `ridge_penalty` can be
+#'   \itemize{
+#'     \item a scalar in which case all parameters except the intercept are penalized.
+#'     \item a vector which has to have the same length as columns in the model matrix
+#'     \item a matrix with the same number of columns as columns in the model matrix. This gives
+#'       maximum flexibility for expert users and allows for full Tikhonov regularization.
+#'   }
+#'   Default: `ridge_penalty = 0`, which is internally replaced with a small positive number for numerical stability.
 #' @param do_cox_reid_adjustment the classical maximum likelihood estimator of the `overdisperion` is biased
 #'   towards small values. McCarthy _et al._ (2012) showed that it is preferable to optimize the Cox-Reid
 #'   adjusted profile likelihood.\cr
@@ -133,9 +144,14 @@
 #'   \item{`size_factors`}{a vector with length `ncol(data)`. The size factors are the inferred
 #'   correction factors for different sizes of each sample. They are also sometimes called the
 #'   exposure factor.}
+#'   \item{`Offset`}{a matrix with the same dimensions as `dim(data)`. If the calculation happened on
+#'   disk, than `Offset` is a `HDF5Matrix`. It contains the `log(size_factors) + offset` from the
+#'   function call.}
 #'   \item{`data`}{a `SummarizedExperiment` that contains the input counts and the `col_data`}
 #'   \item{`model_matrix`}{a matrix with dimensions `ncol(data) x n_coefficients`. It is build based
 #'   on the `design` argument.}
+#'   \item{`design_formula`}{the formula that used to fit the model, or `NULL` otherwise}
+#'   \item{`ridge_penalty`}{a vector with the specification of the ridge penalty.}
 #' }
 #'
 #' @examples
@@ -163,6 +179,9 @@
 #'  fit <- glm_gp(Y, design = ~ fav_food + city + age, col_data = data)
 #'  summary(fit)
 #'
+#'  # Specify 'ridge_penalty' to penalize extreme Beta coefficients
+#'  fit_reg <- glm_gp(Y, design = ~ fav_food + city + age, col_data = data, ridge_penalty = 1.5)
+#'  summary(fit_reg)
 #'
 #'
 #' @seealso [overdispersion_mle()] and [overdispersion_shrinkage()] for the internal functions that do the
@@ -200,6 +219,7 @@ glm_gp <- function(data,
                    size_factors = c("normed_sum", "deconvolution", "poscounts"),
                    overdispersion = TRUE,
                    overdispersion_shrinkage = TRUE,
+                   ridge_penalty = 0,
                    do_cox_reid_adjustment = TRUE,
                    subsample = FALSE,
                    on_disk = NULL,
@@ -231,6 +251,7 @@ glm_gp <- function(data,
               size_factors = size_factors,
               overdispersion = overdispersion,
               overdispersion_shrinkage = overdispersion_shrinkage,
+              ridge_penalty = ridge_penalty,
               do_cox_reid_adjustment = do_cox_reid_adjustment,
               subsample = subsample,
               verbose = verbose)
@@ -246,8 +267,13 @@ glm_gp <- function(data,
   res$design_formula <- des$design_formula
   colnames(res$Beta) <- colnames(res$model_matrix)
   rownames(res$Beta) <- rownames(data)
+  if(! is.null(res$ridge_penalty)){
+    names(res$ridge_penalty) <- colnames(res$model_matrix)
+  }
   rownames(res$Mu) <- rownames(data)
   colnames(res$Mu) <- colnames(data)
+  rownames(res$Offset) <- rownames(data)
+  colnames(res$Offset) <- colnames(data)
   names(res$overdispersions) <- rownames(data)
   names(res$deviances) <- rownames(data)
   names(res$size_factors) <- colnames(data)
@@ -282,15 +308,20 @@ handle_data_parameter <- function(data, on_disk){
   }else if(canCoerce(data, "SummarizedExperiment")){
     se <- as(data, "SummarizedExperiment")
     data_mat <- handle_data_parameter(SummarizedExperiment::assay(se), on_disk)
-  }else if(is(data, "dgCMatrix")) {
-    stop("glmGamPoi does not yet support sparse input data of type 'dgCMatrix'. ",
-         "If your data fits into memory, please cast it to a dense matrix with ",
-         "'as.matrix(data)' or if it is too big, convert it to an on-disk dataset ",
-         "with the 'HDF5Array' package. ")
+  }else if(is(data, "dgCMatrix") || is(data, "dgTMatrix")) {
+    if(isTRUE(on_disk)){
+      data_mat <- HDF5Array::writeHDF5Array(data)
+    }else if(isFALSE(on_disk)){
+      data_mat <- as.matrix(data)
+    }else{
+      stop("glmGamPoi does not yet support sparse input data of type '", class(data),"'. ",
+           "Please explicitly set the 'on_disk' parameter to force a conversion to a dense format either in-memory ('on_disk = FALSE') ",
+           "or on-disk ('on_disk = TRUE')")
+    }
   }else{
-    stop("Cannot handle data of class ", class(data), ".",
-         "It must be of type matrix, DelayedArray, SummarizedExperiment, ",
-         "or something that can be cast to a SummarizedExperiment")
+    stop("Cannot handle data of class '", class(data), "'.",
+         "It must be of type dense matrix object (i.e., a base matrix or DelayedArray),",
+         " or a container for such a matrix (for example: SummarizedExperiment).")
   }
   data_mat
 }
@@ -311,6 +342,8 @@ get_col_data <- function(data, col_data){
 handle_design_parameter <- function(design, data, col_data, reference_level){
   n_samples <- ncol(data)
 
+  ignore_degeneracy <- isTRUE(attr(design, "ignore_degeneracy"))
+
   # Handle the design parameter
   if(is.matrix(design)){
     if(! is.null(reference_level)){
@@ -332,18 +365,22 @@ handle_design_parameter <- function(design, data, col_data, reference_level){
              "the number of samples: ", n_samples)
       }
     }
-    model_matrix <- convert_chr_vec_to_model_matrix(design, reference_level)
-    design_formula <- NULL
+    tmp <- convert_chr_vec_to_model_matrix(design, reference_level)
+    model_matrix <- tmp$model_matrix
+    design_formula <- tmp$formula
+    attr(design_formula, "constructed_from") <- "vector"
   }else if(inherits(design,"formula")){
-    model_matrix <- convert_formula_to_model_matrix(design, col_data, reference_level)
-    design_formula <- design
+    col_data <- relevel_columns_to_reference_level(col_data, reference_level)
+    tmp <- convert_formula_to_model_matrix(design, col_data)
+    model_matrix <- tmp$model_matrix
+    design_formula <- tmp$formula
+    attr(design_formula, "constructed_from") <- "formula"
   }else{
     stop(paste0("design argment of class ", class(design), " is not supported. Please ",
                 "specify a `model_matrix`, a `character vector`, or a `formula`."))
   }
 
-
-  if(nrow(model_matrix) != ncol(data)) stop("Number of rows in col_data does not match number of columns of data")
+  if(nrow(model_matrix) != ncol(data)) stop("Number of rows in col_data does not match number of columns of data.")
   if(! is.null(rownames(model_matrix)) &&
      ! all(rownames(model_matrix) == as.character(seq_len(nrow(model_matrix)))) && # That's the default rownames
      ! is.null(colnames(data))){
@@ -355,35 +392,56 @@ handle_design_parameter <- function(design, data, col_data, reference_level){
         stop("The rownames of the model_matrix / col_data do not match the column names of data.")
       }
     }
-
   }
 
-  if(ncol(model_matrix) >= ncol(data)){
-    stop("The model_matrix:\n", format_matrix(head(model_matrix)), "\nhas more columns (", ncol(model_matrix),
-         ") than the there are samples in the data matrix\n",
-         format_matrix(head(data[,seq_len(min(5, ncol(data))),drop=FALSE])),
-         "\nwhich has ", ncol(data), " columns. ",
-         "Too few replicates / too many coefficients to fit model.")
+  if(any(DelayedMatrixStats::rowAnyNAs(model_matrix))){
+    stop("The design matrix contains 'NA's for sample ",
+         paste0(head(which(DelayedMatrixStats::rowAnyNAs(model_matrix))), collapse = ", "),
+         ". Please remove them before you call 'glm_gp()'.")
+  }
+
+  if(ncol(model_matrix) >= n_samples && ! ignore_degeneracy){
+    stop("The model_matrix has more columns (", ncol(model_matrix),
+         ") than the there are samples in the data matrix (", n_samples, " columns).\n",
+         "Too few replicates / too many coefficients to fit model.\n",
+         "The head of the design matrix: \n", format_matrix(head(model_matrix, n = 3)), "\n",
+         "The head of the data: \n", format_matrix(head(data[,seq_len(min(5, ncol(data))),drop=FALSE], n = 3)))
   }
 
   # Check rank of model_matrix
   qr_mm <- qr(model_matrix)
-  if(qr_mm$rank < ncol(model_matrix) && n_samples > 0){
+  if(qr_mm$rank < ncol(model_matrix) && n_samples > 0  && ! ignore_degeneracy){
     is_zero_column <- DelayedMatrixStats::colCounts(model_matrix, value = 0) == nrow(model_matrix)
     if(any(is_zero_column)){
-      stop("The model matrix:\n", format_matrix(head(model_matrix)), "\nseems degenerate ('matrix_rank(model_matrix) < ncol(model_matrix)'). ",
-           "Column ", paste0(which(is_zero_column), collapse = ", "), " contains only zeros.")
+      stop("The model matrix seems degenerate ('matrix_rank(model_matrix) < ncol(model_matrix)'). ",
+           "Column ", paste0(head(which(is_zero_column), n=10), collapse = ", "), " contains only zeros. \n",
+           "The head of the design matrix: \n", format_matrix(head(model_matrix, n = 3)))
     }else{
-      stop("The model matrix:\n", format_matrix(head(model_matrix)), "\nseems degenerate ('matrix_rank(model_matrix) < ncol(model_matrix)'). ",
-           "Some columns are perfectly collinear. Did you maybe include the same coefficient twice?")
+      stop("The model matrix seems degenerate ('matrix_rank(model_matrix) < ncol(model_matrix)'). ",
+           "Some columns are perfectly collinear. Did you maybe include the same coefficient twice?\n",
+           "The head of the design matrix: \n", format_matrix(head(model_matrix, n = 3)))
     }
   }
 
   rownames(model_matrix) <- colnames(data)
   validate_model_matrix(model_matrix, data)
+  model_matrix <- add_attr_if_intercept(model_matrix)
   list(model_matrix = model_matrix, design_formula = design_formula)
 }
 
+add_attr_if_intercept <- function(model_matrix){
+  intercept_position <- 0
+  for(col_idx in seq_len(ncol(model_matrix))){
+    has_intercept <- all(model_matrix[,col_idx] == 1)
+    if(has_intercept){
+      intercept_position <- col_idx
+      break
+    }
+  }
+  attr(model_matrix, "intercept_position") <- intercept_position
+  model_matrix
+}
+
 
 
 handle_subsample_parameter <- function(data, subsample){
@@ -398,6 +456,60 @@ handle_subsample_parameter <- function(data, subsample){
 }
 
 
+handle_ridge_penalty_parameter <- function(ridge_penalty, model_matrix, verbose){
+  if(! is.null(ridge_penalty)){
+    # This penalty is helpful to protect against numerical instability
+    minimal_penalty <- 1e-10 / nrow(model_matrix)
+    intercept_position <- attr(model_matrix, "intercept_position")
+
+
+    if(length(ridge_penalty) == 1){
+      if(abs(ridge_penalty) < minimal_penalty){
+        ridge_penalty <- minimal_penalty
+      }
+      ridge_target <- attr(ridge_penalty, "target")
+      ridge_penalty <- rep_len(ridge_penalty, ncol(model_matrix))
+      attr(ridge_penalty, "target") <- ridge_target
+      if(! is.null(intercept_position) && intercept_position[1] != 0){
+        ridge_penalty[intercept_position] <- minimal_penalty
+      }
+    }else if(is.matrix(ridge_penalty)){
+      if(ncol(ridge_penalty) != ncol(model_matrix)){
+        stop("'ridge_penalty' is a matrix, but its dimensions do not match ",
+             "the column of the model_matrix (", ncol(model_matrix), ").")
+      }
+      diag(ridge_penalty)[abs(diag(ridge_penalty)) <  minimal_penalty] <- minimal_penalty
+    }else if(length(ridge_penalty) == ncol(model_matrix)){
+      # Got a full length ridge_penalty, check if this conflicts with intercept
+      if(! is.null(intercept_position) && any(abs(ridge_penalty[intercept_position]) > minimal_penalty)){
+        warning("A ridge penalty for each column of the design matrix was provided, including the intercept ",
+                "in column ", intercept_position, ". Are you sure this is correct?\n",
+                "To avoid this message, set the ridge_penalty[", intercept_position, "] to a value ",
+                "smaller than '1e-10/nrow(model_matrix)'.")
+      }
+      ridge_penalty[abs(ridge_penalty) < minimal_penalty] <- minimal_penalty
+    }else{
+      stop("The definition of the ridge penalty does not match the model_matrix. ",
+           "It must either be of length 1 or the number of columns in the design matrix.\n",
+           "If length(ridge_penalty) == 1, it is applied to all columns except the intercept.")
+    }
+
+
+    ridge_target <- attr(ridge_penalty, "target")
+    target_length <- if(is.matrix(ridge_penalty)) ncol(ridge_penalty) else  length(ridge_penalty)
+    if(is.null(ridge_target)){
+      ridge_target <- rep_len(0, target_length)
+    }else if(length(ridge_target) != target_length){
+      stop("Size of ridge_penalty and 'attr(ridge_penalty, \"target\") must correspond.")
+    }
+    attr(ridge_penalty, "target") <- ridge_target
+  }
+
+
+  ridge_penalty
+}
+
+
 
 
 validate_model_matrix <- function(matrix, data){
@@ -406,44 +518,7 @@ validate_model_matrix <- function(matrix, data){
 }
 
 
-
-convert_chr_vec_to_model_matrix <- function(design, reference_level){
-  if(! is.factor(design)){
-    design_fct <- as.factor(design)
-  }else{
-    design_fct <- design
-  }
-
-  if(length(levels(design_fct)) == 1){
-    # All entries are identical build an intercept only model
-    mm <- matrix(1, nrow=length(design_fct), ncol=1)
-    colnames(mm) <- levels(design_fct)
-  }else if(is.null(reference_level)){
-    helper_df <- data.frame(x_ = design_fct)
-    mm <- stats::model.matrix.default(~ x_ - 1, helper_df)
-    colnames(mm) <- sub("^x_", "", colnames(mm))
-  }else{
-    design_fct <- stats::relevel(design_fct, ref = reference_level)
-    helper_df <- data.frame(x_ = design_fct)
-    mm <- stats::model.matrix.default(~ x_ + 1, helper_df)
-    colnames(mm)[-1] <- paste0(sub("^x_", "", colnames(mm)[-1]), "_vs_", reference_level)
-  }
-  colnames(mm)[colnames(mm) == "(Intercept)"] <- "Intercept"
-  mm
-}
-
-extract_data_from_formula <- function(formula, col_data, encl = parent.frame()){
-  if(length(formula) < 3){
-    stop("The formula does not have a left hand side. Please call this function like this:\n",
-         "'glm_gp(data = mat, design = ~ a + b + c, ...)'", call. = FALSE)
-  }
-  data <- eval(formula[[2]], envir = col_data, enclos = encl)
-  formula[[2]] <- NULL
-  list(data = data, design = formula)
-}
-
-
-convert_formula_to_model_matrix <- function(formula, col_data, reference_level=NULL){
+relevel_columns_to_reference_level <- function(col_data, reference_level = NULL){
   if(! is.null(reference_level)){
     has_ref_level <- vapply(col_data, function(x){
       any(!is.na(x) & x == reference_level)
@@ -458,21 +533,74 @@ convert_formula_to_model_matrix <- function(formula, col_data, reference_level=N
       stats::relevel(col, ref = reference_level)
     })
   }
+  col_data
+}
+
+
+convert_chr_vec_to_model_matrix <- function(design, reference_level){
+  if(! is.factor(design)){
+    design_fct <- as.factor(design)
+  }else{
+    design_fct <- design
+  }
+
+  if(length(levels(design_fct)) == 1){
+    helper_df <- data.frame(x_ = design_fct)
+    tmp <- convert_formula_to_model_matrix(~ 1, col_data = helper_df)
+    mm <- tmp$model_matrix
+    formula <- tmp$formula
+    colnames(mm) <- levels(design_fct)
+  }else if(is.null(reference_level)){
+    helper_df <- data.frame(x_ = design_fct)
+    tmp <- convert_formula_to_model_matrix(~ x_ - 1, col_data = helper_df)
+    mm <- tmp$model_matrix
+    formula <- tmp$formula
+    colnames(mm) <- sub("^x_", "", colnames(mm))
+  }else{
+    design_fct <- stats::relevel(design_fct, ref = reference_level)
+    helper_df <- data.frame(x_ = design_fct)
+    tmp <- convert_formula_to_model_matrix(~ x_ + 1, col_data = helper_df)
+    mm <- tmp$model_matrix
+    formula <- tmp$formula
+    colnames(mm)[-1] <- paste0(sub("^x_", "", colnames(mm)[-1]), "_vs_", reference_level)
+  }
+  colnames(mm)[colnames(mm) == "(Intercept)"] <- "Intercept"
+  list(formula = formula, model_matrix = mm)
+}
+
+convert_formula_to_model_matrix <- function(formula, col_data){
+  attr(col_data, "na.action") <- "na.pass"
   tryCatch({
-    mm <- stats::model.matrix.default(formula, col_data)
+    mf <- model.frame(formula, data = col_data, drop.unused.levels = TRUE)
+    terms <- attr(mf, "terms")
+    attr(terms, "xlevels") <- stats::.getXlevels(terms, mf)
+    mm <- stats::model.matrix.default(terms, mf)
   }, error = function(e){
     # Try to extract text from error message
     match <- regmatches(e$message, regexec("object '(.+)' not found", e$message))[[1]]
     if(length(match) == 2){
-      stop("Object '", match[2], "' not found. Allowed variables are:\n",
-           paste0(colnames(col_data), collapse = ", "))
+      stop("Error while parsing the formula (", formula, ").\n",
+           "Variable '", match[2], "' not found in col_data or global environment. Possible variables are:\n",
+           paste0(colnames(col_data), collapse = ", "), call. = FALSE)
     }else{
       stop(e$message)
     }
   })
 
+  # Otherwise every copy of the model stores the whole global environment!
+  attr(terms, ".Environment") <- c()
   colnames(mm)[colnames(mm) == "(Intercept)"] <- "Intercept"
-  mm
+  list(formula = terms, model_matrix = mm)
+}
+
+extract_data_from_formula <- function(formula, col_data, encl = parent.frame()){
+  if(length(formula) < 3){
+    stop("The formula does not have a left hand side. Please call this function like this:\n",
+         "'glm_gp(data = mat, design = ~ a + b + c, ...)'", call. = FALSE)
+  }
+  data <- eval(formula[[2]], envir = col_data, enclos = encl)
+  formula[[2]] <- NULL
+  list(data = data, design = formula)
 }
 
 
